from __future__ import annotations

import random
import string
from typing import Union, Generator, Callable, SupportsRound, Any


class IterationNotCompleted(BaseException):
    """Generator stopped iteration. Make sure it iterates over all arrays length"""
    pass


class UndefinedVariable(Exception):
    """Exception raised for UndefinedVariable"""

    def __init__(self, variable):
        self.message = f"Variable {variable} not defined."


class BoundingError(Exception):
    """Exception raised for BoundingError"""

    def __init__(self, lower_bound, upper_bound):
        self.message = f"Lower bound less than upper bound." + \
                       f" lower bound: {lower_bound}, upper bound: {upper_bound}"
        super(BoundingError, self).__init__(self.message)


class Variables:
    """Simple Base class for all variables

    :kwargs
    -------
        * **generator**: ``Callable[..., Generator]``
        A Callable function that returns a Generator used to generate variable.

        * **decimal_places**: ``int``
        Rounds generated variables, **default**: ``no rounding``
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.generator = kwargs.get('generator')
        self.decimal_places = kwargs.get('decimal_places')

        # last will used to access last generated variable
        # in case of use Variable in other generators as parameters
        self.last = None

        # Initialize
        self.next()

    def rounder(self, val):
        if self.decimal_places and isinstance(val, SupportsRound):
            self.last = round(val, self.decimal_places)
        else:
            self.last = val
        return self.last

    def next(self):
        tmp_args = [x if not isinstance(x, (IntVar, FloatVar)) else x.last for x in self.args]
        self.last = self.rounder(self.generator(*tmp_args))
        return self.last


class BoundedVar(Variables):
    """A simple wrapper for variables that has bounding option.

    :raise *BoundingError* if bounding is wrong.
    """

    def __init__(self, lower_bound, upper_bound, *args, **kwargs):
        tmp_upper = upper_bound if not isinstance(upper_bound, Variables) else upper_bound.last
        tmp_lower = lower_bound if not isinstance(lower_bound, Variables) else lower_bound.last
        if tmp_upper < tmp_lower:
            raise BoundingError(lower_bound, upper_bound)
        super().__init__(lower_bound, upper_bound, *args, **kwargs)


class IntVar(BoundedVar):
    """Generates random random integer between lower and upper bound
    using random.randint callable.
    """

    def __init__(self, lower_bound: Union[IntVar, int], upper_bound: Union[IntVar, int], **kwargs):
        super().__init__(lower_bound, upper_bound, generator=random.randint, **kwargs)


class FloatVar(BoundedVar):
    """Generates random random float between lower and upper bound
    using random.uniform callable.
    """

    def __init__(self, lower_bound: Union[float, int, IntVar, FloatVar],
                 upper_bound: Union[float, int, IntVar, FloatVar], **kwargs):
        super().__init__(lower_bound, upper_bound, generator=random.uniform, **kwargs)


class Collections(Variables):
    """A base class for all collection type variables.
    use this CustomArray() instead if you want to make cus
    """

    def __init__(self, *args, **kwargs):
        self.length = kwargs.get('length') if kwargs.get('length') else 1
        super(Collections, self).__init__(*args, **kwargs)

    def next(self):
        # Using temp args to get current args Variable if they are
        # [:Variable:] for current generation,

        tmp_args = [x if not isinstance(x, (IntVar, FloatVar)) else x.last for x in self.args]
        tmp_length = self.length if not isinstance(self.length, IntVar) else self.length.last

        # not using temp args/length will cause to set arguments as a not
        # changeable integer for next generations.

        return [self.rounder(self.generator(*tmp_args)) for _ in range(tmp_length)]


class CustomArray(Collections):
    """A class to build custom arrays using a Generator."""

    # The difference with :Collections: class is :CustomArray: gets a Callable[..., Generator]
    # that yields each member for a generation, But Collection uses a generator
    # that returns each member of array(e.g random.randint).
    def __init__(self, length: Union[int, IntVar], generator: Callable[..., Generator[Any, Any, Any]], *args, **kwargs):
        super().__init__(*args, generator=generator, length=length, **kwargs)

    def next(self):
        tmp_args = [x if not isinstance(x, (IntVar, FloatVar)) else x.last for x in self.args]
        tmp_length = self.length if not isinstance(self.length, IntVar) else self.length.last

        # Making a generator from Callable[..., Generator] function for each generation
        gen = self.generator(*tmp_args)
        try:
            self.last = [self.rounder(next(gen)) for _ in range(tmp_length)]
        except StopIteration:
            raise IterationNotCompleted("\nGenerator stopped iteration. Make sure it iterates over all arrays length.")
        return self.last


class IntArray(Collections):
    """"Generates random integer for each member of array using random.randint generator"""

    def __init__(self, lower_bound: Union[IntVar, int], upper_bound: Union[IntVar, int],
                 length: Union[IntVar, int]):
        super().__init__(lower_bound, upper_bound, length=length, generator=random.randint, decimal_places=0)


class FloatArray(Collections):
    """"Generates random float for each member of array using random.uniform generator"""

    def __init__(self, lower_bound: Union[int, float, IntVar, FloatVar],
                 upper_bound: Union[int, float, IntVar, FloatVar],
                 length: Union[int, IntVar], decimal_places: int = 1):
        super().__init__(lower_bound, upper_bound, length=length, generator=random.uniform,
                         decimal_places=decimal_places)


class ChoiceList(Collections):
    """Generates random choice from given list with random.choice generator"""

    def __init__(self, length: Union[int, IntVar], choice_list: list, *args, **kwargs):
        super().__init__(choice_list, *args, generator=random.choice, length=length, **kwargs)


class CharArray(ChoiceList):
    """"Generates random choice from all available english characters"""

    def __init__(self, length: Union[int, IntVar]):
        super().__init__(length, string.ascii_letters)



from typing import Protocol, runtime_checkable


@runtime_checkable
class OutputWriter(Protocol):
    """
    The testcase writer interface
    you should Implement these methods to make your own writer.
    """

    def printer(self, inp: str, otp: str, counter: int):
        pass



from .case import Case

from .generator import IntArray
from .generator import IntVar
from .generator import FloatVar
from .generator import FloatArray
from .generator import Collections
from .generator import CharArray
from .generator import ChoiceList
from .generator import Variables
from .generator import CustomArray

from .helper import OutputHelper



from os import path, makedirs


def list_printer(lst, sep="\n", end=""):
    return sep.join([str(x) for x in lst]) + end


def string_printer(lst, end=''):
    return list_printer(lst, "", end)


class OutputHelper:
    def __init__(self, in_prefix: str = "input", out_prefix: str = "output",
                 input_directory: str = "in", output_directory: str = "out",
                 test_directory: str = "tests", zip_name: str = "tests"):
        self.in_prefix = in_prefix
        self.out_prefix = out_prefix
        self.test_directory = test_directory
        self.input_path = path.join(self.test_directory, input_directory)
        self.output_path = path.join(self.test_directory, output_directory)
        self.zip_name = zip_name

    def create_directories(self):
        if not path.exists(self.input_path):
            makedirs(self.input_path)

        if not path.exists(self.output_path):
            makedirs(self.output_path)

    @staticmethod
    def write_to_file(printable: str, dir_path, file_name):
        relative_path = path.join(dir_path, file_name)
        with open(relative_path, "w") as f:
            f.write(printable)

    def printer(self, inp: str, outp: str, _counter: int):
        self.create_directories()
        OutputHelper.write_to_file(inp, self.input_path, f"{self.in_prefix}{str(_counter)}.txt")
        OutputHelper.write_to_file(outp, self.output_path, f"{self.out_prefix}{str(_counter)}.txt")



from setuptools import setup
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pycontest',
    version='1.2',
    packages=['pycontest'],
    url='https://github.com/matinhimself/pycontest',
    license='MIT',
    author='Matin Habibi',
    author_email='aktualled@gmail.com',
    description='An easy to use test case generator.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.8",
)



from unittest import mock
from io import StringIO
from contextlib import redirect_stdout
from typing import Callable, Any, TextIO, Union

from pycontest.helper import OutputHelper
from pycontest.writer import OutputWriter


class NoOtpStrMethod(Exception):
    """Exception raised for NoToStringError"""

    def __init__(self):
        self.message = "Override __otp_str__ method to customize output style."
        super(NoOtpStrMethod, self).__init__(self.message)


class Case:
    """A base class that TestCases will inherit that, by default it will generate a testCase only once
    you can change that by changing the batch_size.
    @param 
    """
    __case_count = 0
    batch_size = 1

    function: Callable[..., Any] = None
    app = None

    input_sequence = None
    output = None

    writer: Union[TextIO, OutputWriter] = OutputHelper()

    separator: str = "\n"

    @staticmethod
    def main(dev_mode=False):

        for sc in Case.__subclasses__():
            tmp = type('tmp', sc.__bases__, dict(sc.__dict__))
            for num in range(sc.batch_size):
                for k, v in vars(sc).items():
                    from pycontest import Variables
                    if isinstance(v, Variables):
                        setattr(tmp, k, v.next())
                tm = tmp()
                tm.config()
                if tm.function and tm.input_sequence is not None:
                    tm.output = tm.function(*tm.input_sequence)
                if tm.app is not None:
                    print(tm.__inp_str__().split('\n'))
                    with mock.patch('builtins.input', side_effect=tm.__inp_str__().split('\n')):
                        with mock.patch('sys.stdout', new=StringIO()) as capturedOutput:
                            exec(tm.app)
                    tm.output = capturedOutput.getvalue()

                tm.printer()
            if dev_mode:
                print(f"{sc.batch_size} tests passed.")
            else:
                print(f"{sc.batch_size} testcase generated.")

    def printer(self):
        if isinstance(self.writer, OutputWriter):
            self.writer.printer(self.__inp_str__(), self.__otp_str__(), Case.__case_count)
            Case.__case_count += 1
            return
        with redirect_stdout(self.writer):
            print(self.__inp_str__() + "\n" + self.__otp_str__() + "\n", end=self.separator)

    def initialize(self, batch_size):
        self.batch_size = batch_size

    def config(self):
        pass

    def __inp_str__(self) -> str:
        return f"{' '.join([str(x) for x in self.input_sequence])}\n" + \
               self.separator

    def __otp_str__(self) -> str:
        if self.output is not None:
            return f"{self.output}"
        else:
            raise NoOtpStrMethod
